/* MemoServ functions.
 *
 * Services is copyright (c) 1996-1998 Andy Church.
 *     E-mail: <achurch@dragonfire.net>
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 */

#include "services.h"
#include "pseudo.h"

const char s_MemoServ[] = MEMOSERV_NAME;


static int delmemo(MemoInfo *mi, int num);

static void do_help(User *u);
static void do_send(User *u);
static void do_list(User *u);
static void do_read(User *u);
static void do_del(User *u);
static void do_set(User *u);
static void do_set_notify(User *u, MemoInfo *mi, char *param);
static void do_set_limit(User *u, MemoInfo *mi, char *param);
static void do_info(User *u);

/*************************************************************************/

static Command cmds[] = {
    { "HELP",       do_help, NULL,  -1,                      -1,-1,-1,-1 },
    { "SEND",       do_send, NULL,  MEMO_HELP_SEND,          -1,-1,-1,-1 },
    { "LIST",       do_list, NULL,  MEMO_HELP_LIST,          -1,-1,-1,-1 },
    { "READ",       do_read, NULL,  MEMO_HELP_READ,          -1,-1,-1,-1 },
    { "DEL",        do_del,  NULL,  MEMO_HELP_DEL,           -1,-1,-1,-1 },
    { "SET",        do_set,  NULL,  MEMO_HELP_SET,           -1,-1,-1,-1 },
    { "SET NOTIFY", NULL,    NULL,  MEMO_HELP_SET_NOTIFY,    -1,-1,-1,-1 },
    { "SET LIMIT",  NULL,    NULL,  -1,                      -1,-1,-1,-1 },
    { "INFO",       do_info, NULL,  -1,
		MEMO_HELP_INFO, MEMO_SERVADMIN_HELP_INFO,
		MEMO_SERVADMIN_HELP_INFO, MEMO_SERVADMIN_HELP_INFO },
    { NULL }
};

/*************************************************************************/
/*************************************************************************/

/* memoserv:  Main MemoServ routine.
 *            Note that the User structure passed to the do_* routines will
 *            always be valid (non-NULL) and will always have a valid
 *            NickInfo pointer in the `ni' field.
 */

void memoserv(const char *source, char *buf)
{
    char *cmd, *s;
    User *u = finduser(source);

    if (!u) {
	log("%s: user record for %s not found", s_MemoServ, source);
	notice(s_MemoServ, u->nick,
		getstring((NickInfo *)NULL, USER_RECORD_NOT_FOUND));
    }

    cmd = strtok(buf, " ");
    if (!cmd) {
	return;
    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_MemoServ, source, "\1PING %s", s);
    } else {
#ifdef SKELETON
	notice_lang(s_MemoServ, u, "%s is currently offline.", s_MemoServ);
#else
	if (!u->ni && stricmp(cmd, "HELP") != 0)
	    notice_lang(s_MemoServ, u, NICK_NOT_REGISTERED_HELP, s_NickServ);
	else
	    run_cmd(s_MemoServ, u, cmds, cmd);
#endif
    }
}

/*************************************************************************/

#ifndef SKELETON

/*************************************************************************/

/* Compatibility memo load routine. */

void load_old_ms_dbase(void)
{
    FILE *f;
    int ver, i, j;
    NickInfo *ni;
    Memo *memos;
    struct memolist_ {
	struct memolist_ *next, *prev;
	char nick[NICKMAX];
	long n_memos;
	Memo *memos;
	long reserved[4];
    } old_memolist;
    struct {
	char sender[NICKMAX];
	long number;
	time_t time;
	char *text;
	short flags;
	short reserved_s;
	long reserved[3];
    } oldmemo;

    if (!(f = open_db(s_MemoServ, "memo.db", "r")))
	return;
    switch (ver = get_file_version(f, "memo.db")) {
      case 4:
      case 3:
      case 2:
      case 1:
	for (i = 33; i < 256; ++i) {
	    while (fgetc(f) == 1) {
		if (1 != fread(&old_memolist, sizeof(old_memolist), 1, f))
		    fatal("Read error on memo.db");
		if (debug >= 3)
		    log("debug: load_old_ms_dbase: got memolist for %s",
				old_memolist.nick);
		old_memolist.memos = memos =
				smalloc(sizeof(Memo) * old_memolist.n_memos);
		for (j = 0; j < old_memolist.n_memos; j++, memos++) {
		    if (1 != fread(&oldmemo, sizeof(oldmemo), 1, f))
			fatal("Read error on memo.db");
		    strscpy(memos->sender, oldmemo.sender, NICKMAX);
		    memos->number = oldmemo.number;
		    memos->time = oldmemo.time;
		    memos->flags = oldmemo.flags;
		}
		memos = old_memolist.memos;
		for (j = 0; j < old_memolist.n_memos; j++) {
		    if (read_string(&memos[j].text, f) < 0)
			fatal("Read error on memo.db");
		}
		ni = findnick(old_memolist.nick);
		if (ni) {
		    ni->memos.memocount = old_memolist.n_memos;
		    ni->memos.memos = old_memolist.memos;
		}
	    }
	}
	break;
      default:
	fatal("Unsupported version number (%d) on memo.db", ver);
    } /* switch (version) */
    close_db(f, "memo.db");
}

/*************************************************************************/

/* check_memos:  See if the given user has any unread memos, and send a
 *               NOTICE to that user if so (and if the appropriate flag is
 *               set).
 */

void check_memos(User *u)
{
    NickInfo *ni;
    int i, newcnt = 0;

    if (!(ni = u->ni) || !(ni->status & (NS_IDENTIFIED | NS_RECOGNIZED)) ||
			 !(ni->flags & NI_MEMO_SIGNON))
	return;

    for (i = 0; i < ni->memos.memocount; i++) {
	if (ni->memos.memos[i].flags & MF_UNREAD)
	    newcnt++;
    }
    if (newcnt > 0) {
	notice_lang(s_MemoServ, u,
		newcnt==1 ? MEMO_HAVE_NEW_MEMO : MEMO_HAVE_NEW_MEMOS, newcnt);
	if (newcnt == 1 && (ni->memos.memos[i-1].flags & MF_UNREAD)) {
	    notice_lang(s_MemoServ, u, MEMO_TYPE_READ_LAST, s_MemoServ);
	} else if (newcnt == 1) {
	    for (i = 0; i < ni->memos.memocount; i++) {
		if (ni->memos.memos[i].flags & MF_UNREAD)
		    break;
	    }
	    notice_lang(s_MemoServ, u, MEMO_TYPE_READ_NUM, s_MemoServ,
			ni->memos.memos[i].number);
	} else {
	    notice_lang(s_MemoServ, u, MEMO_TYPE_LIST_NEW, s_MemoServ);
	}
    }
    if (ni->memos.memomax > 0 && ni->memos.memocount >= ni->memos.memomax) {
	if (ni->memos.memocount > ni->memos.memomax)
	    notice_lang(s_MemoServ, u, MEMO_OVER_LIMIT, ni->memos.memomax);
	else
	    notice_lang(s_MemoServ, u, MEMO_AT_LIMIT, ni->memos.memomax);
    }
}

/*************************************************************************/
/*********************** MemoServ private routines ***********************/
/*************************************************************************/

/* Return the MemoInfo corresponding to the given nick or channel name.
 * Return in `ischan' 1 if the name was a channel name, else 0.
 */

static MemoInfo *getmemoinfo(const char *name, int *ischan)
{
    if (*name == '#') {
	ChannelInfo *ci;
	if (ischan)
	    *ischan = 1;
	if (ci = cs_findchan(name))
	    return &ci->memos;
	else
	    return NULL;
    } else {
	NickInfo *ni;
	if (ischan)
	    *ischan = 0;
	if (ni = findnick(name))
	    return &getlink(ni)->memos;
	else
	    return NULL;
    }
}

/*************************************************************************/

/* Delete a memo by number.  Return 1 if the memo was found, else 0. */

static int delmemo(MemoInfo *mi, int num)
{
    int i;

    for (i = 0; i < mi->memocount; i++) {
	if (mi->memos[i].number == num)
	    break;
    }
    if (i < mi->memocount) {
	free(mi->memos[i].text); /* Deallocate memo text memory */
	mi->memocount--;	 /* One less memo now */
	if (i < mi->memocount)	 /* Move remaining memos down a slot */
	    memmove(mi->memos + i, mi->memos + i+1,
				sizeof(Memo) * (mi->memocount - i));
	if (mi->memocount == 0) { /* If no more memos, free array */
	    free(mi->memos);
	    mi->memos = NULL;
	}
	return 1;
    } else {
	return 0;
    }
}

/*************************************************************************/
/*********************** MemoServ command routines ***********************/
/*************************************************************************/

/* Return a help message. */

static void do_help(User *u)
{
    char *cmd = strtok(NULL, "");

    if (!cmd) {
	notice_help(s_MemoServ, u, MEMO_HELP, s_NickServ);
    } else if (stricmp(cmd, "SET LIMIT") == 0) {
	if (is_services_oper(u))
	    notice_help(s_MemoServ, u, MEMO_SERVADMIN_HELP_SET_LIMIT,
			DEF_MAX_MEMOS);
	else
	    notice_help(s_MemoServ, u, MEMO_HELP_SET_LIMIT, DEF_MAX_MEMOS);
    } else {
	help_cmd(s_MemoServ, u, cmds, cmd);
    }
}

/*************************************************************************/

/* Send a memo to a nick/channel. */

static void do_send(User *u)
{
    char *source = u->nick;
    int ischan;
    MemoInfo *mi;
    Memo *m;
    char *name = strtok(NULL, " ");
    char *text = strtok(NULL, "");
    time_t now = time(NULL);

    if (!text) {
	notice_lang(s_MemoServ, u, SYNTAX_ERROR,
		"\2SEND \37nick\37 \37memo-text\37\2");
	notice_lang(s_MemoServ, u, MORE_INFO, s_MemoServ, "SEND");

    } else if (!(mi = getmemoinfo(name, &ischan))) {
	notice_lang(s_MemoServ, u,
		ischan ? CHAN_X_NOT_REGISTERED : NICK_X_NOT_REGISTERED, name);

#if MEMO_SEND_WAIT > 0
    } else if (u && u->lastmemosend+MEMO_SEND_WAIT > now) {
	u->lastmemosend = now;
	notice_lang(s_MemoServ, u, MEMO_SEND_PLEASE_WAIT, MEMO_SEND_WAIT);
#endif

    } else if (mi->memomax == 0) {
	notice_lang(s_MemoServ, u, MEMO_X_GETS_NO_MEMOS, name);

    } else if (mi->memomax > 0 && mi->memocount >= mi->memomax) {
	notice_lang(s_MemoServ, u, MEMO_X_HAS_TOO_MANY_MEMOS, name);

    } else {
#if MEMO_SEND_WAIT > 0
	u->lastmemosend = now;
#endif
	mi->memocount++;
	mi->memos = srealloc(mi->memos, sizeof(Memo) * mi->memocount);
	m = &mi->memos[mi->memocount-1];
	strscpy(m->sender, source, NICKMAX);
	if (mi->memocount > 1) {
	    m->number = m[-1].number + 1;
	    if (m->number < 1) {
		int i;
		for (i = 0; i < mi->memocount; i++)
		    mi->memos[i].number = i+1;
	    }
	} else {
	    m->number = 1;
	}
	m->time = time(NULL);
	m->text = sstrdup(text);
	m->flags = MF_UNREAD;
	notice_lang(s_MemoServ, u, MEMO_SENT, name);
	if (!ischan) {
	    NickInfo *ni = findnick(name);
	    if ((ni->flags & NI_MEMO_RECEIVE) && (u = finduser(name))) {
		notice_lang(s_MemoServ, u, MEMO_NEW_MEMO_ARRIVED,
			source, s_MemoServ, m->number);
	    }
	}
    }
}

/*************************************************************************/

/* List the memos (if any) for the source nick or given channel. */

static void do_list(User *u)
{
    char *param = strtok(NULL, " "), *chan = NULL;
    ChannelInfo *ci;
    MemoInfo *mi;
    Memo *m;
    int i;
    char timebuf[64];
    time_t t;
    struct tm tm;

    if (param && *param == '#') {
	chan = param;
	param = strtok(NULL, " ");
	if (!(ci = cs_findchan(chan))) {
	    notice_lang(s_MemoServ, u, CHAN_X_NOT_REGISTERED, chan);
	    return;
	} else if (!check_access(u, ci, CA_MEMO)) {
	    notice_lang(s_MemoServ, u, ACCESS_DENIED);
	    return;
	}
	mi = &ci->memos;
    } else {
	mi = &u->ni->memos;
    }
    if (param && stricmp(param, "NEW") != 0) {
	notice_lang(s_MemoServ, u, SYNTAX_ERROR,
		"\2LIST [\37channel\37] [NEW]\2");
	notice_lang(s_MemoServ, u, MORE_INFO, s_MemoServ, "LIST");
    } else if (mi->memocount == 0) {
	if (chan)
	    notice_lang(s_MemoServ, u, MEMO_X_HAS_NO_MEMOS, chan);
	else
	    notice_lang(s_MemoServ, u, MEMO_HAVE_NO_MEMOS);
    } else {
	if (param) {
	    for (i = 0, m = mi->memos; i < mi->memocount; i++, m++) {
		if (m->flags & MF_UNREAD)
		    break;
	    }
	    if (i == mi->memocount) {
		if (chan)
		    notice_lang(s_MemoServ, u, MEMO_X_HAS_NO_NEW_MEMOS, chan);
		else
		    notice_lang(s_MemoServ, u, MEMO_HAVE_NO_NEW_MEMOS);
		return;
	    }
	}
	if (chan) {
	    notice_lang(s_MemoServ, u,
			param ? MEMO_LIST_CHAN_NEW_MEMOS : MEMO_LIST_CHAN_MEMOS,
			chan ? chan : u->nick, s_MemoServ, chan);
	} else {
	    notice_lang(s_MemoServ, u,
			param ? MEMO_LIST_NEW_MEMOS : MEMO_LIST_MEMOS,
			chan ? chan : u->nick, s_MemoServ);
	}
	notice_lang(s_MemoServ, u, MEMO_LIST_HEADER);
	for (i = 0, m = mi->memos; i < mi->memocount; i++, m++) {
	    if (param && !(m->flags & MF_UNREAD))
		continue;
	    t = m->time;
	    tm = *localtime(&t);
	    strftime_lang(timebuf, sizeof(timebuf),
			u, STRFTIME_DATE_TIME_FORMAT, &tm);
	    timebuf[sizeof(timebuf)-1] = 0;	/* just in case */
	    notice_lang(s_MemoServ, u, MEMO_LIST_FORMAT,
			(m->flags & MF_UNREAD) ? '*' : ' ',
			m->number, m->sender, timebuf);
	}

    }
}

/*************************************************************************/

/* Read a memo. */

static void do_read(User *u)
{
    MemoInfo *mi;
    Memo *m;
    ChannelInfo *ci;
    char *numstr = strtok(NULL, " "), *chan = NULL;
    int num;

    if (numstr && *numstr == '#') {
	chan = numstr;
	numstr = strtok(NULL, " ");
	if (!(ci = cs_findchan(chan))) {
	    notice_lang(s_MemoServ, u, CHAN_X_NOT_REGISTERED, chan);
	    return;
	} else if (!check_access(u, ci, CA_MEMO)) {
	    notice_lang(s_MemoServ, u, ACCESS_DENIED);
	    return;
	}
	mi = &ci->memos;
    } else {
	mi = &u->ni->memos;
    }
    num = numstr ? atoi(numstr) : -1;
    if (!numstr || (stricmp(numstr,"LAST") != 0 && num <= 0)) {
	notice_lang(s_MemoServ, u, SYNTAX_ERROR,
		"\2READ [\37channel\37] {\37num\37 | LAST}");
	notice_lang(s_MemoServ, u, MORE_INFO, s_MemoServ, "READ");

    } else if (mi->memocount == 0) {
	if (chan)
	    notice_lang(s_MemoServ, u, MEMO_X_HAS_NO_MEMOS, chan);
	else
	    notice_lang(s_MemoServ, u, MEMO_HAVE_NO_MEMOS);

    } else {
	int i;

	if (stricmp(numstr, "LAST") == 0) {
	    for (i = 0; i < mi->memocount-1; i++)
		;
	} else {
	    for (i = 0; i < mi->memocount; i++) {
		if (mi->memos[i].number == num)
		    break;
	    }
	}
	if (i >= mi->memocount) {
	    notice_lang(s_MemoServ, u, MEMO_DOES_NOT_EXIST, num);
	} else {
	    char timebuf[64];
	    time_t t;
	    struct tm tm;

	    m = &mi->memos[i];
	    t = m->time;
	    tm = *localtime(&t);
	    strftime_lang(timebuf, sizeof(timebuf),
			u, STRFTIME_DATE_TIME_FORMAT, &tm);
	    timebuf[sizeof(timebuf)-1] = 0;
	    if (chan)
		notice_lang(s_MemoServ, u, MEMO_CHAN_HEADER, m->number,
			m->sender, timebuf, s_MemoServ, chan, m->number);
	    else
		notice_lang(s_MemoServ, u, MEMO_HEADER, m->number,
			m->sender, timebuf, s_MemoServ, m->number);
	    notice_lang(s_MemoServ, u, MEMO_TEXT, m->text);
	    m->flags &= ~MF_UNREAD;
	}

    }
}

/*************************************************************************/

/* Delete memos. */

static void do_del(User *u)
{
    MemoInfo *mi;
    ChannelInfo *ci;
    char *numstr = strtok(NULL, ""), *chan = NULL, *s;
    int num, num2, last, last0, i;
    char buf[BUFSIZE], *end = buf;
    int delcount = 0;

    if (numstr && *numstr == '#') {
	chan = strtok(numstr, " ");
	numstr = strtok(NULL, "");
	if (!(ci = cs_findchan(chan))) {
	    notice_lang(s_MemoServ, u, CHAN_X_NOT_REGISTERED, chan);
	    return;
	} else if (!check_access(u, ci, CA_MEMO)) {
	    notice_lang(s_MemoServ, u, ACCESS_DENIED);
	    return;
	}
	mi = &ci->memos;
    } else {
	mi = &u->ni->memos;
    }
    if (!numstr || (!isdigit(*numstr) && stricmp(numstr, "ALL") != 0)) {
	notice_lang(s_MemoServ, u, SYNTAX_ERROR,
		"\2DEL [\37channel\37] {\37num\37[-\37num\37][,\37num\37...] | ALL}");
	notice_lang(s_MemoServ, u, MORE_INFO, s_MemoServ, "DEL");
    } else if (mi->memocount == 0) {
	if (chan)
	    notice_lang(s_MemoServ, u, MEMO_HAVE_NO_MEMOS);
	else
	    notice_lang(s_MemoServ, u, MEMO_HAVE_NO_MEMOS);
    } else {
	if (isdigit(*numstr)) {
	    /* Delete a specific memo or memos. */
	    num = strtol(numstr, &s, 10);
	    if (!*s) {
		/* Handle the special case of a single number specially. */
		if (delmemo(mi, num))
		    notice_lang(s_MemoServ, u, MEMO_DELETED_ONE, num);
		else
		    notice_lang(s_MemoServ, u, MEMO_DOES_NOT_EXIST, num);
		return;
	    }
	    numstr = s;
	    last0 = -1;  /* Beginning of range of last memos deleted */
	    last = -1;   /* Last memo deleted */
	    /*
	     * This algorithm ignores invalid characters, ignores a dash
	     * when it precedes a comma, and ignores everything from the
	     * end of a valid number or range to the next comma or null.
	     */
	    for (;;) {
		numstr += strcspn(numstr, "0123456789,-");
		num2 = num;
		if (*numstr == '-') {
		    numstr++;
		    numstr += strcspn(numstr, "0123456789,");
		    if (isdigit(*numstr))
			num2 = strtol(numstr, &numstr, 10);
		}
		for (i = num; i <= num2; i++) {
		    if (delmemo(mi, i)) {
			delcount++;
			if (i != last+1) {
			    if (last != -1) {
				if (last0 != last)
				    end += snprintf(end, sizeof(buf)-(end-buf),
						",%d-%d", last0, last);
				else
				    end += snprintf(end, sizeof(buf)-(end-buf),
						",%d", last);
			    }
			    last0 = i;
			}
			last = i;
		    }
		}
		numstr += strcspn(numstr, ",");
		if (*numstr) {
		    numstr++;
		    num = strtol(numstr, &numstr, 10);
		    numstr += strcspn(numstr, "0123456789,-");
		} else {
		    break;
		}
	    }
	    if (last != -1) {
		/* Some memos got deleted; tell them which ones. */
		if (delcount > 1) {
		    if (last0 != last)
			end += snprintf(end, sizeof(buf)-(end-buf),
				",%d-%d", last0, last);
		    else
			end += snprintf(end, sizeof(buf)-(end-buf),
				",%d", last);
		    /* "buf+1" here because *buf == ',' */
		    notice_lang(s_MemoServ, u, MEMO_DELETED_SEVERAL, buf+1);
		} else {
		    notice_lang(s_MemoServ, u, MEMO_DELETED_ONE, last);
		}
	    } else {
		/* No memos were deleted.  Tell them so. */
		notice_lang(s_MemoServ, u, MEMO_DELETED_NONE);
	    }
	} else {
	    /* Delete all memos. */
	    for (i = 0; i < mi->memocount; i++)
		free(mi->memos[i].text);
	    free(mi->memos);
	    mi->memos = NULL;
	    mi->memocount = 0;
	    notice_lang(s_MemoServ, u, MEMO_DELETED_ALL);
	}
    }
}

/*************************************************************************/

static void do_set(User *u)
{
    char *cmd    = strtok(NULL, " ");
    char *param  = strtok(NULL, "");
    MemoInfo *mi = &u->ni->memos;

    if (readonly) {
	notice_lang(s_MemoServ, u, MEMO_SET_DISABLED);
	return;
    }
    if (!param) {
	notice_lang(s_MemoServ, u, SYNTAX_ERROR,
		"\2SET \37option\37 \37parameters\37\2");
	notice_lang(s_MemoServ, u, MORE_INFO, s_MemoServ, "SET");
    } else if (stricmp(cmd, "NOTIFY") == 0) {
	do_set_notify(u, mi, param);
    } else if (stricmp(cmd, "LIMIT") == 0) {
	do_set_limit(u, mi, param);
    } else {
	notice_lang(s_MemoServ, u, MEMO_SET_UNKNOWN_OPTION, strupper(cmd));
	notice_lang(s_MemoServ, u, MORE_INFO, s_MemoServ, "SET");
    }
}

/*************************************************************************/

static void do_set_notify(User *u, MemoInfo *mi, char *param)
{
    if (stricmp(param, "ON") == 0) {
	u->ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE;
	notice_lang(s_MemoServ, u, MEMO_SET_NOTIFY_ON, s_MemoServ);
    } else if (stricmp(param, "LOGON") == 0) {
	u->ni->flags |= NI_MEMO_SIGNON;
	u->ni->flags &= ~NI_MEMO_RECEIVE;
	notice_lang(s_MemoServ, u, MEMO_SET_NOTIFY_LOGON, s_MemoServ);
    } else if (stricmp(param, "NEW") == 0) {
	u->ni->flags &= ~NI_MEMO_SIGNON;
	u->ni->flags |= NI_MEMO_RECEIVE;
	notice_lang(s_MemoServ, u, MEMO_SET_NOTIFY_NEW, s_MemoServ);
    } else if (stricmp(param, "OFF") == 0) {
	u->ni->flags &= ~(NI_MEMO_SIGNON | NI_MEMO_RECEIVE);
	notice_lang(s_MemoServ, u, MEMO_SET_NOTIFY_OFF, s_MemoServ);
    } else {
	notice_lang(s_MemoServ, u, SYNTAX_ERROR,
		"\2SET NOTIFY {ON | LOGON | NEW | OFF}\2");
	notice_lang(s_MemoServ, u, MORE_INFO, s_MemoServ, "SET NOTIFY");
    }
}

/*************************************************************************/

static void do_set_limit(User *u, MemoInfo *mi, char *param)
{
    char *p1 = strtok(param, " ");
    char *p2 = strtok(NULL, " ");
    char *p3 = strtok(NULL, " ");
    char *user = NULL, *chan = NULL;
    int32 limit;
    NickInfo *ni = u->ni;
    ChannelInfo *ci = NULL;
    int is_servadmin = is_services_admin(u);

    if (p1 && *p1 == '#') {
	chan = p1;
	p1 = p2;
	p2 = p3;
	p3 = strtok(NULL, " ");
	if (!(ci = cs_findchan(chan))) {
	    notice_lang(s_MemoServ, u, CHAN_X_NOT_REGISTERED, chan);
	    return;
	} else if (!is_servadmin && !check_access(u, ci, CA_MEMO)) {
	    notice_lang(s_MemoServ, u, ACCESS_DENIED);
	    return;
	}
	mi = &ci->memos;
    }
    if (is_servadmin) {
	if (p2 && stricmp(p2, "HARD") != 0 && !chan) {
	    if (!(ni = findnick(p1))) {
		notice_lang(s_MemoServ, u, NICK_X_NOT_REGISTERED, p1);
		return;
	    }
	    ni = getlink(ni);
	    user = p1;
	    mi = &ni->memos;
	    p1 = p2;
	    p2 = p3;
	} else if (!p1) {
	    notice_lang(s_MemoServ, u, SYNTAX_ERROR,
		    "\2SET LIMIT [\37user\37 | \37channel\37] {\37limit\37 | NONE} [HARD]\2");
	    notice_lang(s_MemoServ, u, MORE_INFO, s_MemoServ, "SET LIMIT");
	    return;
	}
	if ((!isdigit(*p1) && stricmp(p1, "NONE") != 0) ||
			(p2 && stricmp(p2, "HARD") != 0)) {
	    notice_lang(s_MemoServ, u, SYNTAX_ERROR,
		    "\2SET LIMIT [\37user\37 | \37channel\37] {\37limit\37 | NONE} [HARD]\2");
	    notice_lang(s_MemoServ, u, MORE_INFO, s_MemoServ, "SET LIMIT");
	    return;
	}
	if (chan) {
	    if (p2)
		ci->flags |= CI_MEMO_HARDMAX;
	    else
		ci->flags &= ~CI_MEMO_HARDMAX;
	} else {
	    if (p2)
		ni->flags |= NI_MEMO_HARDMAX;
	    else
		ni->flags &= ~NI_MEMO_HARDMAX;
	}
	limit = atoi(p1);
	if (limit < 0 || limit > 32767) {
	    notice_lang(s_MemoServ, u, MEMO_SET_LIMIT_OVERFLOW, 32767);
	    limit = 32767;
	}
	if (stricmp(p1, "NONE") == 0)
	    limit = -1;
    } else {
	if (!p1 || p2 || !isdigit(*p1)) {
	    notice_lang(s_MemoServ, u, SYNTAX_ERROR,
		    "\2SET LIMIT [\37channel\37] \37limit\37\2");
	    notice_lang(s_MemoServ, u, MORE_INFO, s_MemoServ, "SET LIMIT");
	    return;
	}
	if (chan && (ci->flags & CI_MEMO_HARDMAX)) {
	    notice_lang(s_MemoServ, u, MEMO_SET_LIMIT_FORBIDDEN, chan);
	    return;
	} else if (!chan && (ni->flags & NI_MEMO_HARDMAX)) {
	    notice_lang(s_MemoServ, u, MEMO_SET_YOUR_LIMIT_FORBIDDEN);
	    return;
	}
	limit = atoi(p1);
	/* The first character is a digit, but we could still go negative
	 * from overflow... watch out! */
	if (limit < 0 || (DEF_MAX_MEMOS > 0 && limit > DEF_MAX_MEMOS)) {
	    if (chan) {
		notice_lang(s_MemoServ, u, MEMO_SET_LIMIT_TOO_HIGH,
			chan, DEF_MAX_MEMOS);
	    } else {
		notice_lang(s_MemoServ, u, MEMO_SET_YOUR_LIMIT_TOO_HIGH,
			DEF_MAX_MEMOS);
	    }
	    return;
	} else if (limit > 32767) {
	    notice_lang(s_MemoServ, u, MEMO_SET_LIMIT_OVERFLOW, 32767);
	    limit = 32767;
	}
    }
    mi->memomax = limit;
    if (limit > 0) {
	if (!chan && ni == u->ni)
	    notice_lang(s_MemoServ, u, MEMO_SET_YOUR_LIMIT, limit);
	else
	    notice_lang(s_MemoServ, u, MEMO_SET_LIMIT,
			chan ? chan : user, limit);
    } else if (limit == 0) {
	if (!chan && ni == u->ni)
	    notice_lang(s_MemoServ, u, MEMO_SET_YOUR_LIMIT_ZERO);
	else
	    notice_lang(s_MemoServ, u, MEMO_SET_LIMIT_ZERO, chan ? chan : user);
    } else {
	if (!chan && ni == u->ni)
	    notice_lang(s_MemoServ, u, MEMO_UNSET_YOUR_LIMIT);
	else
	    notice_lang(s_MemoServ, u, MEMO_UNSET_LIMIT, chan ? chan : user);
    }
}

/*************************************************************************/

static void do_info(User *u)
{
    MemoInfo *mi;
    NickInfo *ni = NULL;
    ChannelInfo *ci = NULL;
    char *name = strtok(NULL, " ");
    int is_servadmin = is_services_admin(u);
    int hardmax = 0;

    if (is_servadmin && name && *name != '#') {
	ni = findnick(name);
	if (!ni) {
	    notice_lang(s_MemoServ, u, NICK_X_NOT_REGISTERED, name);
	    return;
	}
	ni = getlink(ni);
	mi = &ni->memos;
	hardmax = ni->flags & NI_MEMO_HARDMAX ? 1 : 0;
    } else if (name && *name == '#') {
	ci = cs_findchan(name);
	if (!ci) {
	    notice_lang(s_MemoServ, u, CHAN_X_NOT_REGISTERED, name);
	    return;
	} else if (!check_access(u, ci, CA_MEMO)) {
	    notice_lang(s_MemoServ, u, ACCESS_DENIED);
	    return;
	}
	mi = &ci->memos;
	hardmax = ci->flags & CI_MEMO_HARDMAX ? 1 : 0;
    } else { /* !name */
	mi = &u->ni->memos;
    }

    if (name && ni != u->ni) {

	if (!mi->memocount) {
	    notice_lang(s_MemoServ, u, MEMO_INFO_X_NO_MEMOS, name);
	} else if (mi->memocount == 1) {
	    if (mi->memos[0].flags & MF_UNREAD)
		notice_lang(s_MemoServ, u, MEMO_INFO_X_MEMO_UNREAD, name);
	    else
		notice_lang(s_MemoServ, u, MEMO_INFO_X_MEMO, name);
	} else {
	    int count = 0, i;
	    for (i = 0; i < mi->memocount; i++) {
		if (mi->memos[i].flags & MF_UNREAD)
		    count++;
	    }
	    if (count == mi->memocount)
		notice_lang(s_MemoServ, u, MEMO_INFO_X_MEMOS_ALL_UNREAD,
			name, count);
	    else if (count == 0)
		notice_lang(s_MemoServ, u, MEMO_INFO_X_MEMOS,
			name, mi->memocount);
	    else if (count == 0)
		notice_lang(s_MemoServ, u, MEMO_INFO_X_MEMOS_ONE_UNREAD,
			name, mi->memocount);
	    else
		notice_lang(s_MemoServ, u, MEMO_INFO_X_MEMOS_SOME_UNREAD,
			name, mi->memocount, count);
	}
	if (mi->memomax >= 0) {
	    if (hardmax)
		notice_lang(s_MemoServ, u, MEMO_INFO_X_HARD_LIMIT,
			name, mi->memomax);
	    else
		notice_lang(s_MemoServ, u, MEMO_INFO_X_LIMIT,
			name, mi->memomax);
	} else {
	    notice_lang(s_MemoServ, u, MEMO_INFO_X_NO_LIMIT, name);
	}

    } else { /* !name || ni == u->ni */

	if (!mi->memocount) {
	    notice_lang(s_MemoServ, u, MEMO_INFO_NO_MEMOS);
	} else if (mi->memocount == 1) {
	    if (mi->memos[0].flags & MF_UNREAD)
		notice_lang(s_MemoServ, u, MEMO_INFO_MEMO_UNREAD);
	    else
		notice_lang(s_MemoServ, u, MEMO_INFO_MEMO);
	} else {
	    int count = 0, i;
	    for (i = 0; i < mi->memocount; i++) {
		if (mi->memos[i].flags & MF_UNREAD)
		    count++;
	    }
	    if (count == mi->memocount)
		notice_lang(s_MemoServ, u, MEMO_INFO_MEMOS_ALL_UNREAD, count);
	    else if (count == 0)
		notice_lang(s_MemoServ, u, MEMO_INFO_MEMOS, mi->memocount);
	    else if (count == 1)
		notice_lang(s_MemoServ, u, MEMO_INFO_MEMOS_ONE_UNREAD,
			mi->memocount);
	    else
		notice_lang(s_MemoServ, u, MEMO_INFO_MEMOS_SOME_UNREAD,
			mi->memocount, count);
	}
	if (mi->memomax == 0) {
	    if (!is_servadmin && hardmax)
		notice_lang(s_MemoServ, u, MEMO_INFO_HARD_LIMIT_ZERO);
	    else
		notice_lang(s_MemoServ, u, MEMO_INFO_LIMIT_ZERO);
	} else if (mi->memomax > 0) {
	    if (!is_servadmin && hardmax)
		notice_lang(s_MemoServ, u, MEMO_INFO_HARD_LIMIT, mi->memomax);
	    else
		notice_lang(s_MemoServ, u, MEMO_INFO_LIMIT, mi->memomax);
	} else {
	    notice_lang(s_MemoServ, u, MEMO_INFO_NO_LIMIT);
	}

    } /* if (name && ni != u->ni) */
}

/*************************************************************************/

#endif	/* !SKELETON */
